-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use work.sha3_pkg.all;
use work.groestl_pkg.all;

-- Groestl fsm2 is responsible for Groestl computation processing
-- Possible generics values:
-- hs = {HASH_SIZE_256, HASH_SIZE_512}

entity groestl_fsm2 is
	generic (hs : integer:=HASH_SIZE_256; ppl : integer := 2);
	port (
		clk 					: in std_logic;
		rst 					: in std_logic;
		
		smsg 					: out std_logic_vector(log2(ppl)-1 downto 0);
		schain					: out std_logic_vector(log2(ppl)-1 downto 0);
		load_hash 				: out std_logic_vector(ppl-1 downto 0);
		
		final					: out std_logic;
		init1					: out std_logic;
		init2					: out std_logic;
		init3					: out std_logic;  
		init4					: out std_logic;
		load_ctr				: out std_logic;
		wr_ctr					: out std_logic;
		wr_result				: out std_logic_vector(ppl-1 downto 0);
		wr_state				: out std_logic;
		
		-- control				   
			--fsm1 hand shake signals
		block_ready_clr 		: out std_logic_vector(ppl-1 downto 0);		
		msg_end_clr 			: out std_logic_vector(ppl-1 downto 0);
		
		block_ready				: in std_logic_vector(ppl-1 downto 0);
		msg_end 				: in std_logic_vector(ppl-1 downto 0);	
				
			--fsm3 handshake signals
		output_busy_set 		: out std_logic_vector(ppl-1 downto 0);
		output_busy				: in  std_logic_vector(ppl-1 downto 0)
	);
end groestl_fsm2;


architecture beh of groestl_fsm2 is
	function get_roundnr ( hs : integer ) return integer is
	begin
		if hs = 256 then
			return 10;
		elsif hs = 512 then
			return 14;
		end if;
	end function get_roundnr ;
	constant roundnr 		: integer := get_roundnr( hs );  -- Add an additional round for block finalization
	constant log2roundnr	: integer := log2( roundnr );
	constant log2roundnrzeros : std_logic_vector(log2roundnr-1 downto 0) := (others => '0') ;
	
	signal pc : std_logic_vector(log2roundnr-1 downto 0);
	signal ziroundnr, ziround0, ei, li : std_logic;
	signal output_data_s : std_logic_vector(ppl-1 downto 0);
	
	signal ppl_stage_cnt : std_logic_vector(log2(ppl)-1 downto 0);
	
	signal sf_set, sf_clr, sf_sig : std_logic_vector(ppl-1 downto 0);
    signal msg_done_clr, msg_done_set, msg_done_sig : std_logic_vector(ppl-1 downto 0);
	signal fin_done_clr, fin_done_set, fin_done_sig : std_logic_vector(ppl-1 downto 0);
	signal snb_sig : std_logic;
	signal block_ready_true: std_logic_vector(PPL-1 downto 0);
	
	signal final_in, init1_in, init2_in : std_logic_vector(ppl-1 downto 0);
	
	constant log2pplzeros : std_logic_vector(log2(ppl)-1 downto 0) := (others =>'0');
	constant log2pplones : std_logic_vector(log2(ppl)-1 downto 0) := (others =>'1');
	
	type smsg_type is array ( 0 to ppl-2 ) of std_logic_vector(log2(ppl)-1 downto 0);
	signal smsg_reg : smsg_type;
begin
	-- Internals
	RoundCounterGen : 
        countern 
        generic map ( N => log2roundnr ) 
        port map ( clk => clk, rst => rst, load => li, en => ei, input => log2roundnrzeros, output => pc);
	
	PPLStageCntGen: process ( clk )
	begin
		if rising_edge( clk ) then
		
			if rst = '1' or ei = '1' or li = '1' then
				ppl_stage_cnt <= (others => '0');
			else
				ppl_stage_cnt <= ppl_stage_cnt + 1;
			end if;
		end if;
	end process;
	
	SmsgGen : process(clk)
	begin
		if rising_edge( clk ) then
			smsg_reg(0) <= ppl_stage_cnt;	 
			if ( ppl > 2 ) then
				for i in 1 to ppl-2 loop
					smsg_reg(i) <= smsg_reg(i-1);
				end loop;	   
			end if;
		end if;
	end process;	
	smsg <= ppl_stage_cnt;
	schain <= smsg_reg(ppl-2);
		
	ziroundnr <= '1' when pc = conv_std_logic_vector(roundnr-1,log2roundnr) else '0';			 	
	ziround0 <= '1' when pc = 0 else '0';
	li <= '1' when (ziroundnr = '1' and ppl_stage_cnt = log2pplones) else '0';
	ei <= '1' when (ppl_stage_cnt = log2pplones) else '0';

	FlagGen: for i in 0 to ppl-1 generate	 				
		SelectFirstReg : sr_reg 
			generic map ( init => '1' )
			port map ( rst => rst, clk => clk, set => sf_set(i), clr => sf_clr(i), output => sf_sig(i));	
		
		MsgDoneReg : sr_reg 
			generic map ( init => '0' )
			port map ( rst => rst, clk => clk, set => msg_done_set(i), clr => msg_done_clr(i), output => msg_done_sig(i));
			
		FinalDoneReg: sr_reg	  -- finalization finished flag
			generic map ( init => '0' )
			port map ( rst => rst, clk => clk, set => fin_done_set(i), clr => fin_done_clr(i), output => fin_done_sig(i));
		
        sf_set(i) 		<= '1' when (ziroundnr = '1' and ppl_stage_cnt = conv_std_logic_vector((i),log2(ppl)) and fin_done_sig(i) = '1') else '0';		
		--sf_clr(i) 		<= '1' when (ziround0 = '1'   and ppl_stage_cnt = conv_std_logic_vector((i),log2(ppl)) and sf_sig(i) = '1' and block_ready_true(i) = '1') else '0';        
		sf_clr(i) 		<= '1' when (snb_sig = '1'   and smsg_reg(ppl-2) = conv_std_logic_vector((i),log2(ppl)) and sf_sig(i) = '1' and block_ready_true(i) = '1') else '0';        
		output_data_s(i)<= '1' when (ziround0 = '1'   and ppl_stage_cnt = conv_std_logic_vector((i),log2(ppl)) and block_ready_true(i) = '1' and msg_done_sig(i) = '0') else '0';
		msg_done_set(i) <= '1' when (ziround0 = '1'   and ppl_stage_cnt = conv_std_logic_vector((i),log2(ppl)) and 
									 msg_end(i) = '1'and block_ready_true(i) = '1' and fin_done_sig(i) = '0' and msg_done_sig(i) = '0' ) else '0';
        msg_done_clr(i) <= '1' when (ziround0 = '1'   and ppl_stage_cnt = conv_std_logic_vector((i),log2(ppl)) and msg_done_sig(i) = '1') else '0';
		fin_done_set(i) <= '1' when (ziround0 = '1'   and ppl_stage_cnt = conv_std_logic_vector((i),log2(ppl)) and msg_done_sig(i) = '1') else '0';
		fin_done_clr(i) <= '1' when (snb_sig = '1' and smsg_reg(ppl-2) = conv_std_logic_vector((i),log2(ppl)) and fin_done_sig(i) = '1') else '0';
					   
	--	wr_result(i) <= '1' when (snb_sig = '1' and smsg_reg(ppl-2) = conv_std_logic_vector((i),log2(ppl)))  or 
--									(sf_sig(i) = '1' and ppl_stage_cnt = conv_std_logic_vector(i,log2(ppl))) else '0';		
		wr_result(i) <= '1' when (snb_sig = '1' and smsg_reg(ppl-2) = conv_std_logic_vector((i),log2(ppl))) else '0';

		final_in(i) <= '1' when (ziroundnr = '1' and fin_done_sig(i) = '1' and ppl_stage_cnt = conv_std_logic_vector(i,log2(ppl))) else '0';
		init1_in(i) <= '1' when (ziround0 = '1' and fin_done_sig(i) = '0' and msg_done_sig(i) = '0' and ppl_stage_cnt = conv_std_logic_vector(i,log2(ppl))) else '0';		
	end generate;																															   
	init2_in <= fin_done_set;
    block_ready_clr <= output_data_s;	
	msg_end_clr		<= output_data_s;
	output_busy_set	<= fin_done_clr;
	load_hash		<= fin_done_clr;	   
	
    snb_sig <= '1' when (ziroundnr = '1' and ppl_stage_cnt = ppl-1) or 
						(ziround0 = '1' and ppl_stage_cnt /= ppl-1 ) else '0';


	BlockReadyTrueReg : process ( clk ) 	-- Allow everything to be synchronize to round number
	begin
		if rising_edge( clk ) then	 
			if rst = '1' then
				block_ready_true <= (others => '0');				
			elsif (pc = roundnr-2 and ppl_stage_cnt = ppl-1) then  				
				block_ready_true <= block_ready;
			end if;
		end if;
	end process;
	
	
	wr_state <= '1';
	load_ctr <= '1' when (ziroundnr = '1' and ppl_stage_cnt = PPL-1) else '0';
	wr_ctr    <= '1' when (ppl_stage_cnt = PPL-1) else '0';	
															
	Ppl2Gen:
		if PPL = 2 generate
			init1 <= init1_in(0) when ppl_stage_cnt = 0 else init1_in(1); 
			
			init2 <= init2_in(0) when ppl_stage_cnt = 0 else init2_in(1); 
			
			init3 <= sf_sig(0) when smsg_reg(ppl-2) = 0 else sf_sig(1);  
			
			init4 <= sf_sig(0) when ppl_stage_cnt = 0 else sf_sig(1); 
			
			final <= final_in(0) when ppl_stage_cnt = 0 else final_in(1); 				
		end generate;
	Ppl4Gen:
		if PPL = 4 generate				 
			with ppl_stage_cnt(1 downto 0) select
			init1 <= init1_in(0) when "00", init1_in(1) when "01", init1_in(2) when "10", init1_in(3) when others;
			
			with ppl_stage_cnt(1 downto 0) select
			init2 <= init2_in(0) when "00", init2_in(1) when "01", init2_in(2) when "10", init2_in(3) when others; 
			
			with smsg_reg(ppl-2)(1 downto 0) select
			init3 <= sf_sig(0) when "00", sf_sig(1) when "01", sf_sig(2) when "10", sf_sig(3) when others; 
			
			with ppl_stage_cnt(1 downto 0) select
			init4 <= sf_sig(0) when "00", sf_sig(1) when "01", sf_sig(2) when "10", sf_sig(3) when others; 
			
			with ppl_stage_cnt(1 downto 0) select
			final <= final_in(0) when "00", final_in(1) when "01", final_in(2) when "10", final_in(3) when others;
		end generate;	
	
	
	
	--final <= '1' when (cstate = finalization ) and (ziroundnr = '1') else '0';
--	init1 <= '1' when  (cstate = idle and block_ready = '1') or
--						((cstate = process_data or cstate = process_last_data) and pc = 0) or
--						(cstate = output_data and output_busy = '0' and block_ready = '1') else '0';
--	init2 <= '1' when (cstate = finalization and pc = 0)  else '0';
--
--	init3 <= '1' when (cstate = reset) or (cstate = write_output) or (cstate = output_data) else '0';
--
--	load_ctr <= '1' when (cstate = reset) or ((cstate = process_data or cstate = process_last_data) and ziroundnr = '1') or
--						(cstate = finalization and pc=roundnr-1) else '0';
--
--	wr_ctr <= '1' when 	(cstate = idle and block_ready = '1') or
--						(cstate = process_data or cstate = process_last_data or cstate = finalization ) else '0';
--	wr_result <= '1' when (cstate = reset) or
--						((cstate = process_data or cstate = process_last_data) and ( ziroundnr = '1' )) or
--						(cstate = finalization and ziroundnr = '1' ) or
--						(cstate = write_output and output_busy = '0') or
--						(cstate = output_data and output_busy = '0')  else '0';
--	wr_state <= '1' when (cstate = idle and block_ready = '1') or
--						(cstate = process_data and ziroundnr = '0') or
--						((cstate = process_data) and ziroundnr = '1' and ((msg_end = '0' and block_ready = '1') or (msg_end = '1'))) or
--						(cstate = process_last_data) or
--						(cstate = finalization and ziroundnr = '0') or
--						(cstate = finalization and ziroundnr = '1' and output_busy = '0') or
--						(cstate = output_data and output_busy = '0') else '0';
end beh;